/**
 * Developed by: The Open Systems Lab
 *             University of Illinois at Urbana-Champaign
 *             Department of Computer Science
 *             Urbana, IL 61801
 *             http://osl.cs.uiuc.edu
 *
 * Contact: http://osl.cs.uiuc.edu/soter
 *
 * Copyright (c) 1998-2011
 * The University of Illinois Board of Trustees.
 *    All Rights Reserved.
 * 
 * Distributed under license: http://osl.cs.uiuc.edu/soter/LICENSE
 * 
 */
package edu.illinois.soter.analysis.scala;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IField;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.classLoader.ShrikeCTMethod;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.callgraph.Entrypoint;
import com.ibm.wala.ipa.callgraph.propagation.PointerKey;
import com.ibm.wala.ipa.summaries.BypassMethodTargetSelector;
import com.ibm.wala.ipa.summaries.XMLMethodSummaryReader;
import com.ibm.wala.shrikeBT.IInstruction;
import com.ibm.wala.shrikeBT.InvokeInstruction;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.SSAAbstractInvokeInstruction;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.types.MethodReference;
import com.ibm.wala.types.Selector;

import edu.illinois.soter.Debugger;
import edu.illinois.soter.analysis.AnalysisHelper;
import edu.illinois.soter.analysis.FieldSensitiveEntrypoint;
import edu.illinois.soter.analysis.OwnershipTransferAnalysis;
import edu.illinois.soter.messagedata.MessageInvocation;

/**
 * 
 * @author Stas Negara
 * 
 */
public class ScalaAnalysis extends OwnershipTransferAnalysis {

	private static final boolean IS_CONSIDERING_ACTOR_CREATION_SITES = Boolean.FALSE;

	private static final String ABSENT_MAIN_ENTRYPOINT = "NONE";

	private static final Selector ACT_METHOD_SELECTOR = Selector.make("act()V");
	private static final Selector APPLY_METHOD_SELECTOR = Selector.make("apply()Ljava/lang/Object;");
	private static final Selector MAIN_METHOD_SELECTOR = Selector.make("main([Ljava/lang/String;)V");

	protected static final String BANG_SIGNATURE_SUFFIX = ".$bang(Ljava/lang/Object;)V";
	protected static final String BANG_QMARK_SIGNATURE_SUFFIX = ".$bang$qmark(Ljava/lang/Object;)Ljava/lang/Object;";
	protected static final String START_SIGNATURE_SUFFIX = ".start()Lscala/actors/Actor;";
	protected static final String EXIT_SIGNATURE_SUFFIX = ".exit(Lscala/actors/AbstractActor;Ljava/lang/Object;)V";

	private static final Set<String> predefinedScalaActors = new HashSet<String>(7);
	private static final Set<String> predefinedScalaActorFactoryUsers = new HashSet<String>(1); //Occured in scala-2.7.7.final

	private final String entryClassName;

	private final Set<String> foundActors = new HashSet<String>();

	static {
		predefinedScalaActors.add("Lscala/actors/Actor$$anon$1");
		predefinedScalaActors.add("Lscala/actors/Actor$$anon$2");
		predefinedScalaActors.add("Lscala/actors/Actor$$anon$3");
		predefinedScalaActors.add("Lscala/actors/Actor$$anon$5");
		predefinedScalaActors.add("Lscala/actors/Futures$$anon$1");
		predefinedScalaActors.add("Lscala/actors/ActorProxy");
		predefinedScalaActors.add("Lscala/actors/remote/DelegateActor");

		predefinedScalaActorFactoryUsers.add("Lscala/actors/Futures$");
	}

	public ScalaAnalysis(String classPath, String entryClassName) {
		super("Lscala/actors/Actor", classPath);
		this.entryClassName = entryClassName;
	}

	@Override
	protected void setupIgnoredFields() {
	}

	@Override
	protected boolean isIgnoredField(IField field) {
		return false;
	}

	@Override
	protected void populateEntrypoints(final Set<Entrypoint> entrypoints) {
		addActorEntrypoints(entrypoints);
		addActorFactoryEntrypoints(entrypoints);
		addMainMethodEntrypoint(entrypoints);
	}

	private void addActorEntrypoints(final Set<Entrypoint> entrypoints) {
		Collection<IClass> actorImplementors = classHierarchy.getImplementors(basicActorReference);
		for (IClass klass : actorImplementors) {
			//Do not consider abstract classes and predefined Scala actor classes
			String className = klass.getName().toString();
			if (!klass.isAbstract() && !predefinedScalaActors.contains(className)) {
				foundActors.add(className);
				IMethod actMethod = lookupMethod(klass, ACT_METHOD_SELECTOR);
				Debugger.debugEntrypointMethod(actMethod);
				entrypoints.add(new FieldSensitiveEntrypoint(actMethod, classHierarchy));
			}
		}
	}

	private void addActorFactoryEntrypoints(final Set<Entrypoint> entrypoints) {
		Iterator<IClass> classIterator = classHierarchy.iterator();
		while (classIterator.hasNext()) {
			IClass klass = classIterator.next();
			if (!predefinedScalaActorFactoryUsers.contains(klass.getName().toString())) {
				for (IMethod method : klass.getDeclaredMethods()) {
					IInstruction[] instructions = AnalysisHelper.getShrikeInstructions((ShrikeCTMethod) method);
					if (instructions != null) {
						for (int i = 0; i < instructions.length; i++) {
							IInstruction instruction = instructions[i];
							if (instruction instanceof InvokeInstruction && isActorFactoryInvocation((InvokeInstruction) instruction)) {
								String closureTypeName = getClosureTypeName(instructions[i - 1]);
								foundActors.add(closureTypeName);
								IMethod applyMethod = lookupMethod(closureTypeName, APPLY_METHOD_SELECTOR);
								Debugger.debugEntrypointMethod(applyMethod);
								entrypoints.add(new FieldSensitiveEntrypoint(applyMethod, classHierarchy));
							}
						}
					}
				}
			}
		}
	}

	private void addMainMethodEntrypoint(final Set<Entrypoint> entrypoints) {
		if (!entryClassName.equals(ABSENT_MAIN_ENTRYPOINT)) {
			String entryTypeName = "L" + entryClassName.replace('.', '/');
			IMethod mainMethod = lookupMethod(entryTypeName, MAIN_METHOD_SELECTOR);
			Debugger.debugEntrypointMethod(mainMethod);
			entrypoints.add(new FieldSensitiveEntrypoint(mainMethod, classHierarchy));
		}
	}

	@Override
	protected BypassMethodTargetSelector getCustomBypassMethodTargetSelector(XMLMethodSummaryReader summary) {
		return new ScalaBypassMethodTargetSelector(classHierarchy.lookupClass(basicActorReference), analysisOptions.getMethodTargetSelector(), summary.getSummaries(), summary.getIgnoredPackages(),
				classHierarchy);
	}

	@Override
	protected boolean isNodeConsideredForMessageInvocations(CGNode cgNode) {
		return true; //consider all nodes
	}

	@Override
	protected int getMessageArgumentsIndex(MethodReference methodReference) {
		if (isMessagePassingSite(methodReference.getSignature())) {
			return 1;
		}
		if (IS_CONSIDERING_ACTOR_CREATION_SITES && isCreateActorSite(methodReference)) {
			return 0;
		}
		return -1;
	}

	private boolean isMessagePassingSite(String methodSignature) {
		return methodSignature.endsWith(BANG_SIGNATURE_SUFFIX) || methodSignature.endsWith(BANG_QMARK_SIGNATURE_SUFFIX);
	}

	private boolean isCreateActorSite(MethodReference methodReference) {
		return AnalysisHelper.isInitMethod(methodReference) && foundActors.contains(methodReference.getDeclaringClass().getName().toString());
	}

	@Override
	protected void addMessageInvocationArguments(MessageInvocation messageInvocation, int messagePassingIndex, int passedPointerValueNumber) {
		CGNode cgNode = messageInvocation.getCGNode();
		IR ir = cgNode.getIR();
		SSAAbstractInvokeInstruction messagePassingInstruction = (SSAAbstractInvokeInstruction) ir.getInstructions()[messagePassingIndex];
		if (isCreateActorSite(messagePassingInstruction.getDeclaredTarget())) {
			addInvokeInstructionArguments(messageInvocation, messagePassingInstruction, messagePassingIndex);
		} else { //message passing invocation
			PointerKey passedPointer = heapModel.getPointerKeyForLocal(cgNode, passedPointerValueNumber);
			InvokeInstructionPair initInvocationPair = getInitInvocationPair(ir, passedPointerValueNumber);
			if (initInvocationPair != null && isUsedOnlyForMessagePassing(ir, passedPointerValueNumber)) {
				messageInvocation.addArgument(messagePassingIndex, getImmediatelyPointedItems(passedPointer));
				addInvokeInstructionArguments(messageInvocation, initInvocationPair.invokeInstruction, initInvocationPair.invokeInstructionBytecodeIndex);
			} else {
				messageInvocation.addArgument(messagePassingIndex, computeTransitiveClosureOfPointedItems(passedPointer));
			}
		}
	}

	private void addInvokeInstructionArguments(MessageInvocation messageInvocation, SSAAbstractInvokeInstruction invokeInstruction, int instructionIndex) {
		for (int i = 1; i < invokeInstruction.getNumberOfUses(); i++) {
			int useValueNumber = invokeInstruction.getUse(i);
			PointerKey usePointer = heapModel.getPointerKeyForLocal(messageInvocation.getCGNode(), useValueNumber);
			//In the current CodeTransformer all transformations would point to the same bytecode index
			messageInvocation.addArgument(instructionIndex, computeTransitiveClosureOfPointedItems(usePointer));
		}
	}

	private Set<Object> getImmediatelyPointedItems(PointerKey pointer) {
		Set<Object> immediatelyPointedItems = new HashSet<Object>();
		Iterator<Object> immediatelyPointedItemsIterator = getImmediatelyPointedItemsIterator(pointer);
		while (immediatelyPointedItemsIterator.hasNext()) {
			immediatelyPointedItems.add(immediatelyPointedItemsIterator.next());
		}
		return immediatelyPointedItems;
	}

	private InvokeInstructionPair getInitInvocationPair(IR ir, int passedPointerValueNumber) {
		Set<SSAInstruction> useInstructions = AnalysisHelper.getUseInstructions(ir, passedPointerValueNumber);
		for (SSAInstruction instruction : useInstructions) {
			if (instruction instanceof SSAAbstractInvokeInstruction) {
				SSAAbstractInvokeInstruction invokeInstruction = (SSAAbstractInvokeInstruction) instruction;
				if (AnalysisHelper.isInitInvocation(invokeInstruction) && invokeInstruction.getUse(0) == passedPointerValueNumber) {
					return new InvokeInstructionPair(invokeInstruction, AnalysisHelper.getBytecodeIndexForCallSite(ir, invokeInstruction.getCallSite()));
				}
			}
		}
		return null; //not found
	}

	private boolean isUsedOnlyForMessagePassing(IR ir, int passedPointerValueNumber) {
		Set<SSAInstruction> useInstructions = AnalysisHelper.getUseInstructions(ir, passedPointerValueNumber);
		for (SSAInstruction instruction : useInstructions) {
			if (instruction instanceof SSAAbstractInvokeInstruction) {
				SSAAbstractInvokeInstruction invokeInstruction = (SSAAbstractInvokeInstruction) instruction;
				for (int i = 0; i < invokeInstruction.getNumberOfUses(); i++) {
					if (invokeInstruction.getUse(i) == passedPointerValueNumber) {
						//Check if it is passed in a message or initialized in this invocation. 
						if (!(i == 1 && isMessagePassingSite(invokeInstruction.getDeclaredTarget().getSignature())) && !(AnalysisHelper.isInitInvocation(invokeInstruction) && i == 0)) {
							return false;
						}
					}
				}
			} else {
				return false;
			}
		}
		return true;
	}

	private boolean isActorFactoryInvocation(InvokeInstruction instruction) {
		return isInvocationOfFactoryMethod(instruction, "actor") || isInvocationOfFactoryMethod(instruction, "link");
		//TODO: Implement when encountered in a Scala program		
		//|| isInvocationOfFactoryMethod(instruction, "reactor");
	}

	private boolean isInvocationOfFactoryMethod(InvokeInstruction instruction, String factoryMethodName) {
		return instruction.getInvocationModeString().equals("VIRTUAL") && instruction.getClassType().equals("Lscala/actors/Actor$;") && instruction.getMethodName().equals(factoryMethodName)
				&& instruction.getMethodSignature().equals("(Lscala/Function0;)Lscala/actors/Actor;");
	}

	private String getClosureTypeName(IInstruction instruction) {
		if (!(instruction instanceof InvokeInstruction)) {
			throw new RuntimeException("Could not get closure from a non invocation instruction: " + instruction);
		}
		InvokeInstruction invokeInstruction = (InvokeInstruction) instruction;
		if (!AnalysisHelper.isInitInvocation(invokeInstruction)) {
			throw new RuntimeException("Could not get closure from a non initialization instruction: " + invokeInstruction);
		}
		String classType = invokeInstruction.getClassType();
		return classType.substring(0, classType.length() - 1); //remove trailing ';'
	}

	/**
	 * A pair of SSAAbstractInvokeInstruction and its bytecode index
	 * 
	 */
	private class InvokeInstructionPair {

		private SSAAbstractInvokeInstruction invokeInstruction;
		private int invokeInstructionBytecodeIndex;

		private InvokeInstructionPair(SSAAbstractInvokeInstruction invokeInstruction, int invokeInstructionBytecodeIndex) {
			this.invokeInstruction = invokeInstruction;
			this.invokeInstructionBytecodeIndex = invokeInstructionBytecodeIndex;
		}
	}

}
